<!DOCTYPE gui SYSTEM "kpartgui.dtd">
<gui name="kdevcppcheck" version="9" translationDomain="kdevcppcheck">
<MenuBar>
  <Menu name="code">
    <Menu name="analyze_file">
      <Action name="cppcheck_file"/>
    </Menu>
    <Menu name="analyze_project">
      <Action name="cppcheck_project"/>
    </Menu>
  </Menu>
</MenuBar>
</gui>
